
import javax.media.opengl.GL;

/**
 * @author Jeremy Smith
 */
public class CS752_Surface 
{
    private final int subdivision = 4; //how surface is divided by control points
    private final int resolution = 20; //how many verticies are between control points
    private final int detail = ((subdivision - 1) * resolution) + 1; //number of total verticies
    
    //basis matrix for Bezier curved surface
    private final float[][] B_Bez = {{-1,  3, -3, 1},
                                     { 3, -6,  3, 0},
                                     {-3,  3,  0, 0},
                                     { 1,  0,  0, 0}};
    
    private float[][][] controlPoints;
    private float[][][] verticies;
    private int[] selectedPoint = {0, 0}; //points to currently selected control point
    
    public CS752_Surface()
    {
        // initalize control points
        controlPoints = new float[subdivision][subdivision][3];
        
        for(int s = 0; s < subdivision; s++)
        {
            for (int t = 0; t < subdivision; t++)
            {
                controlPoints[s][t][0] = (float)s/(subdivision - 1);
                controlPoints[s][t][1] = (float)t/(subdivision - 1);
                controlPoints[s][t][2] = 0.0f;
            }
        }
        
        // initialize verticies
        verticies = new float[detail][detail][3];
        
        for(int s = 0; s < detail; s++)
        {
            for (int t = 0; t < detail; t++)
            {
                verticies[s][t][0] = (float)s/(detail - 1);
                verticies[s][t][1] = (float)t/(detail - 1);
                verticies[s][t][2] = 0.0f;
            }
        }
    }
    
    public void selectNextPoint()
    {
        selectedPoint[0]++;
        
        if(selectedPoint[0] >= subdivision)
        {
            selectedPoint[0] = 0;
            selectedPoint[1]++;
            
            if(selectedPoint[1] >= subdivision)
                selectedPoint[1] = 0;
        }
    }
    
    public void movePoint(float adjustX, float adjustY, float adjustZ)
    {
        controlPoints[selectedPoint[0]][selectedPoint[1]][0] += adjustX;
        controlPoints[selectedPoint[0]][selectedPoint[1]][1] += adjustY;
        controlPoints[selectedPoint[0]][selectedPoint[1]][2] += adjustZ;
    }
    
    public void draw(GL gl, int shaderprogram)
    {
        gl.glPushMatrix();
        
        //disable shader and lighting for point drawing
        gl.glUseProgram(0); 
        gl.glDisable(GL.GL_LIGHTING);
        
        //display selected control point
        gl.glColor3f(0,1,0);
        gl.glPointSize(10.0f);
        gl.glBegin(GL.GL_POINTS);
        
        gl.glVertex3fv(controlPoints[selectedPoint[0]][selectedPoint[1]], 0);
        
        gl.glEnd();
        
        //display control points
        gl.glColor3f(0,0,1);
        gl.glPointSize(5.0f);
        gl.glBegin(GL.GL_POINTS);
        
        for(int s = 0; s < subdivision; s++)
        {
            for (int t = 0; t < subdivision; t++)
            {
                gl.glVertex3fv(controlPoints[s][t], 0);
            }
        }
        
        gl.glEnd();
        
        //reenable shader and lighting for surface drawing
        gl.glUseProgram(shaderprogram); 
        gl.glEnable(GL.GL_LIGHTING);

        //display surface
        gl.glBegin(GL.GL_TRIANGLE_STRIP);
        for(int s = 0; s < detail-1; s++)
        {
            gl.glBegin(GL.GL_TRIANGLE_STRIP);
            for (int t = 0; t < detail; t++)
            {
                gl.glTexCoord2f((float)s/(detail - 1), (float)t/(detail - 1));
                gl.glVertex3fv(verticies[s][t], 0);
                
                gl.glTexCoord2f((float)(s+1)/(detail - 1), (float)t/(detail - 1));
                gl.glVertex3fv(verticies[s+1][t], 0);
                
            }
            gl.glEnd();
        }
        
        gl.glPopMatrix();
    }
    
    public void calculateVerticies()
    {
        //      [ p_00 p_01 p_02 p_03 ]
        //  G = [ p_10 p_11 p_12 p_13 ]
        //      [ p_20 p_21 p_22 p_23 ]
        //      [ p_30 p_31 p_32 p_33 ]
        
        float[][] G_x = new float[subdivision][subdivision];
        float[][] G_y = new float[subdivision][subdivision];
        float[][] G_z = new float[subdivision][subdivision];
        
        for(int s = 0; s < subdivision; s++)
        {
            for (int t = 0; t < subdivision; t++)
            {
                G_x[s][t] = controlPoints[s][t][0];
                G_y[s][t] = controlPoints[s][t][1];
                G_z[s][t] = controlPoints[s][t][2];
            }
        }
        
        // C = B_Bez * G * B_Bez^T
        float[][] C_x = multiplyMatrixByMatrix(B_Bez, multiplyMatrixByMatrix(G_x, B_Bez));
        float[][] C_y = multiplyMatrixByMatrix(B_Bez, multiplyMatrixByMatrix(G_y, B_Bez));
        float[][] C_z = multiplyMatrixByMatrix(B_Bez, multiplyMatrixByMatrix(G_z, B_Bez));
        
        //v(s,t) = s * C * t^T
        for(int s = 0; s < detail; s++)
        {
            for (int t = 0; t < detail; t++)
            {
                float [] sVec = new float[4];
                float [] tVec = new float[4];
                
                float sVar = (float)s/(detail - 1);
                float tVar = (float)t/(detail - 1);
                
                sVec[0] = sVar * sVar * sVar;
                sVec[1] = sVar * sVar;
                sVec[2] = sVar;
                sVec[3] = 1.0f;
                
                tVec[0] = tVar * tVar * tVar;
                tVec[1] = tVar * tVar;
                tVec[2] = tVar;
                tVec[3] = 1.0f;
                
                verticies[s][t][0] = multiplyVectorByVector(multiplyVectorByMatrix(sVec, C_x), tVec);
                verticies[s][t][1] = multiplyVectorByVector(multiplyVectorByMatrix(sVec, C_y), tVec);
                verticies[s][t][2] = multiplyVectorByVector(multiplyVectorByMatrix(sVec, C_z), tVec);
            }
        }
    }
    
    // Multiply two 4x4 matricies
    private float[][] multiplyMatrixByMatrix(float[][] matrixA, float[][] matrixB)
    {
        float[][] result = new float[4][4];
        
        for(int i = 0; i < 4; i++)
        {
            for(int j = 0; j < 4; j++)
            {
                float sum = 0.0f;   
                for(int k = 0; k < 4; k++)
                {
                    sum += matrixA[i][k] * matrixB[k][j];
                }
                result[i][j] = sum;
            }
        }
        
        return result;
    }

    // Mutliply a 4x1 vector with a 4x4 matrix
    private float[] multiplyVectorByMatrix(float[] vectorA, float[][] matrixB)
    {
        float[] result = new float[4];
        
        for(int i = 0; i < 4; i++)
        {
            float sum = 0.0f;
            for(int j = 0; j < 4; j++)
            {
                sum += vectorA[j] * matrixB[j][i];
            }
            result[i] = sum;
        }
        
        return result;
    }

    // Multiply two 4x1 vectors
    private float multiplyVectorByVector(float[] vectorA, float[] vectorB) 
    {
        return (vectorA[0]*vectorB[0]) + (vectorA[1]*vectorB[1]) + (vectorA[2]*vectorB[2]) + (vectorA[3]*vectorB[3]);
    }
}
